from metaflow.plugins.datastores.s3_storage import S3Storage
from typing import TYPE_CHECKING
import uuid

import os
from metaflow.plugins.storage_executor import (
    StorageExecutor,
    handle_executor_exceptions,
)

from itertools import starmap


from metaflow.plugins.datatools.s3.s3 import (
    S3,
    S3Client,
    S3PutObject,
    check_s3_deps,
    MetaflowS3NotFound,
)
from metaflow.metaflow_config import DATASTORE_SYSROOT_S3, ARTIFACT_LOCALROOT, TEMPDIR
from metaflow.datastore.datastore_storage import CloseAfterUse, DataStoreStorage

import tempfile
from concurrent.futures import as_completed
import shutil


try:
    # python2
    from urlparse import urlparse
except:
    # python3
    from urllib.parse import urlparse


class S3CompatibleStorage(S3Storage):

    TYPE = "s3-compatible"

    @check_s3_deps
    def __init__(self, root=None, session_vars=None, role_arn=None, client_params=None):
        super(S3Storage, self).__init__(root)
        self._s3_session_vars = session_vars
        self._s3_role_arn = role_arn
        self._s3_client_params = client_params
        self.s3_client = None

    @classmethod
    def get_datastore_root_from_config(cls, echo, create_on_absent=True):
        # TODO : Figure if this is the right way or not.
        return DATASTORE_SYSROOT_S3

    def save_files(self, key_path_tuples, overwrite=False):
        with S3(
            s3root=self.datastore_root,
            role=self._s3_role_arn,
            session_vars=self._s3_session_vars,
            client_params=self._s3_client_params,
        ) as s3:
            s3.put_files(
                [
                    S3PutObject(
                        key=key,
                        path=path,
                    )
                    for key, path in key_path_tuples
                ],
                overwrite=overwrite,
            )

    def save_file(self, key, path, overwrite=False):
        with S3(
            s3root=self.datastore_root,
            role=self._s3_role_arn,
            session_vars=self._s3_session_vars,
            client_params=self._s3_client_params,
        ) as s3:
            s3.put_files(
                [
                    S3PutObject(
                        key=key,
                        path=path,
                    )
                ],
                overwrite=overwrite,
            )

    def load_files(self, keys):
        if len(keys) == 0:
            return CloseAfterUse(iter([]))

        s3 = S3(
            s3root=self.datastore_root,
            role=self._s3_role_arn,
            session_vars=self._s3_session_vars,
            client_params=self._s3_client_params,
        )

        def iter_results():
            # We similarly do things in parallel for many files. This is again
            # a hack.
            if len(keys) > 10:
                results = s3.get_many(keys, return_missing=True, return_info=True)
                for r in results:
                    if r.exists:
                        yield r.key, r.path, r.metadata
                    else:
                        yield r.key, None, None
            else:
                for p in keys:
                    r = s3.get(p, return_missing=True, return_info=True)
                    if r.exists:
                        yield r.key, r.path, r.metadata
                    else:
                        yield r.key, None, None

        return CloseAfterUse(iter_results(), closer=s3)

    def is_file(self, paths):
        with S3(
            s3root=self.datastore_root,
            tmproot=ARTIFACT_LOCALROOT,
            role=self._s3_role_arn,
            session_vars=self._s3_session_vars,
            client_params=self._s3_client_params,
        ) as s3:
            if len(paths) > 10:
                s3objs = s3.info_many(paths, return_missing=True)
                return [s3obj.exists for s3obj in s3objs]
            else:
                result = []
                for path in paths:
                    result.append(s3.info(path, return_missing=True).exists)
                return result

    def info_file(self, path):
        with S3(
            s3root=self.datastore_root,
            tmproot=ARTIFACT_LOCALROOT,
            role=self._s3_role_arn,
            session_vars=self._s3_session_vars,
            client_params=self._s3_client_params,
        ) as s3:
            s3obj = s3.info(path, return_missing=True)
            return s3obj.exists, s3obj.metadata

    def size_file(self, path):
        with S3(
            s3root=self.datastore_root,
            tmproot=ARTIFACT_LOCALROOT,
            role=self._s3_role_arn,
            session_vars=self._s3_session_vars,
            client_params=self._s3_client_params,
        ) as s3:
            s3obj = s3.info(path, return_missing=True)
            return s3obj.size

    def list_content(self, paths):
        strip_prefix_len = len(self.datastore_root.rstrip("/")) + 1
        with S3(
            s3root=self.datastore_root,
            tmproot=ARTIFACT_LOCALROOT,
            role=self._s3_role_arn,
            session_vars=self._s3_session_vars,
            client_params=self._s3_client_params,
        ) as s3:
            results = s3.list_paths(paths)
            return [
                self.list_content_result(
                    path=o.url[strip_prefix_len:], is_file=o.exists
                )
                for o in results
            ]

    def save_bytes(self, path_and_bytes_iter, overwrite=False, len_hint=0):
        def _convert():
            # Output format is the same as what is needed for S3PutObject:
            # key, value, path, content_type, metadata
            for path, obj in path_and_bytes_iter:
                if isinstance(obj, tuple):
                    yield path, obj[0], None, None, obj[1]
                else:
                    yield path, obj, None, None, None

        with S3(
            s3root=self.datastore_root,
            tmproot=ARTIFACT_LOCALROOT,
            role=self._s3_role_arn,
            session_vars=self._s3_session_vars,
            client_params=self._s3_client_params,
        ) as s3:
            # HACK: The S3 datatools we rely on does not currently do a good job
            # determining if uploading things in parallel is more efficient than
            # serially. We use a heuristic for now where if we have a lot of
            # files, we will go in parallel and if we have few files, we will
            # serially upload them. This is not ideal because there is also a size
            # factor and one very large file with a few other small files, for
            # example, would benefit from a parallel upload.
            #
            # In the case of save_artifacts, currently len_hint is based on the
            # total number of artifacts, not taking into account how many of them
            # already exist in the CAS, i.e. it can be a gross overestimate. As a
            # result, it is possible we take a latency hit by using put_many only
            # for a few artifacts.
            #
            # A better approach would be to e.g. write all blobs to temp files
            # and based on the total size and number of files use either put_files
            # (which avoids re-writing the files) or s3.put sequentially.
            if len_hint > 10:
                # Use put_many
                s3.put_many(starmap(S3PutObject, _convert()), overwrite)
            else:
                # Sequential upload
                for key, obj, _, _, metadata in _convert():
                    s3.put(key, obj, overwrite=overwrite, metadata=metadata)

    def load_bytes(self, paths):
        if len(paths) == 0:
            return CloseAfterUse(iter([]))

        s3 = S3(
            s3root=self.datastore_root,
            tmproot=ARTIFACT_LOCALROOT,
            role=self._s3_role_arn,
            session_vars=self._s3_session_vars,
            client_params=self._s3_client_params,
        )

        def iter_results():
            # We similarly do things in parallel for many files. This is again
            # a hack.
            if len(paths) > 10:
                results = s3.get_many(paths, return_missing=True, return_info=True)
                for r in results:
                    if r.exists:
                        yield r.key, r.path, r.metadata
                    else:
                        yield r.key, None, None
            else:
                for p in paths:
                    r = s3.get(p, return_missing=True, return_info=True)
                    if r.exists:
                        yield r.key, r.path, r.metadata
                    else:
                        yield r.key, None, None

        return CloseAfterUse(iter_results(), closer=s3)


def _load_bytes_single_cw(
    role_arn, session_vars, client_params, dir_path, s3_path, _key
):
    from boto3.session import Session
    from botocore.exceptions import ClientError
    from botocore.config import Config

    session = Session()
    if client_params is None:
        client_params = {}

    _client_params = client_params.copy()
    if _client_params.get("config") and type(_client_params["config"]) == dict:
        _client_params["config"] = Config(**_client_params["config"])

    client = session.client("s3", **_client_params)
    bucket = urlparse(s3_path).netloc
    key = urlparse(s3_path).path.lstrip("/")
    tmp_filename = os.path.join(dir_path, str(uuid.uuid4()))
    try:
        client.download_file(Bucket=bucket, Key=key, Filename=tmp_filename)
    except ClientError as e:
        if e.response["Error"]["Code"] == "404":
            return _key, None, None
        else:
            raise
    return _key, tmp_filename, None


class CoreweaveStorage(S3CompatibleStorage):
    TYPE = "coreweave"

    def __init__(self, root=None, session_vars=None, role_arn=None, client_params=None):
        super(CoreweaveStorage, self).__init__(
            root, session_vars, role_arn, client_params
        )
        from metaflow import get_aws_client

        self._get_aws_client = get_aws_client
        # Executor needs to be thread based since boto3 will have issues in a processbased executor.
        self._executor = StorageExecutor(use_processes=False)

    @handle_executor_exceptions
    def load_bytes(self, paths, temp_dir_root=None):
        if len(paths) == 0:
            return CloseAfterUse(iter([]))

        if temp_dir_root is None:
            temp_dir_root = ARTIFACT_LOCALROOT

        tmpdir = tempfile.mkdtemp(
            suffix=None,
            prefix="metaflow.coreweave.load_bytes.",
            dir=temp_dir_root,
        )
        full_paths = [os.path.join(self.datastore_root, key) for key in paths]

        try:
            futures = [
                self._executor.submit(
                    _load_bytes_single_cw,
                    self._s3_role_arn,
                    self._s3_session_vars,
                    self._s3_client_params,
                    tmpdir,
                    s3_path,
                    key,
                )
                for s3_path, key in zip(full_paths, paths)
            ]

            items = [future.result() for future in as_completed(futures)]
        except Exception:
            if os.path.exists(tmpdir):
                shutil.rmtree(tmpdir)
            raise

        class _Closer(object):
            @staticmethod
            def close():
                if os.path.isdir(tmpdir):
                    shutil.rmtree(tmpdir)

        return CloseAfterUse(iter(items), closer=_Closer)

    def load_files(self, keys):
        return self.load_bytes(keys, temp_dir_root=TEMPDIR)



from metaflow.decorators import FlowDecorator, StepDecorator
from metaflow.metadata_provider import MetaDatum
from metaflow.exception import MetaflowException


class ArtifactStoreFlowDecorator(FlowDecorator):
    """
    Allows setting external datastores to save data for the
    `@checkpoint`/`@model`/`@huggingface_hub` decorators.

    This decorator is useful when users wish to save data to a different datastore
    than what is configured in Metaflow. This can be for variety of reasons:

    1. Data security: The objects needs to be stored in a bucket (object storage) that is not accessible by other flows.
    2. Data Locality: The location where the task is executing is not located in the same region as the datastore.
        - Example: Metaflow datastore lives in US East, but the task is executing in Finland datacenters.
    3. Data Lifecycle Policies: The objects need to be archived / managed separately from the Metaflow managed objects.
        - Example: Flow is training very large models that need to be stored separately and will be deleted more aggressively than the Metaflow managed objects.

    Usage:
    ----------

    - Using a custom IAM role to access the datastore.

        ```python
        @with_artifact_store(
            type="s3",
            config=lambda: {
                "root": "s3://my-bucket-foo/path/to/root",
                "role_arn": ROLE,
            },
        )
        class MyFlow(FlowSpec):

            @checkpoint
            @step
            def start(self):
                with open("my_file.txt", "w") as f:
                    f.write("Hello, World!")
                self.external_bucket_checkpoint = current.checkpoint.save("my_file.txt")
                self.next(self.end)

        ```

    - Using credentials to access the s3-compatible datastore.

        ```python
        @with_artifact_store(
            type="s3",
            config=lambda: {
                "root": "s3://my-bucket-foo/path/to/root",
                "client_params": {
                    "aws_access_key_id": os.environ.get("MY_CUSTOM_ACCESS_KEY"),
                    "aws_secret_access_key": os.environ.get("MY_CUSTOM_SECRET_KEY"),
                },
            },
        )
        class MyFlow(FlowSpec):

            @checkpoint
            @step
            def start(self):
                with open("my_file.txt", "w") as f:
                    f.write("Hello, World!")
                self.external_bucket_checkpoint = current.checkpoint.save("my_file.txt")
                self.next(self.end)

        ```

    - Accessing objects stored in external datastores after task execution.

        ```python
        run = Run("CheckpointsTestsFlow/8992")
        with artifact_store_from(run=run, config={
            "client_params": {
                "aws_access_key_id": os.environ.get("MY_CUSTOM_ACCESS_KEY"),
                "aws_secret_access_key": os.environ.get("MY_CUSTOM_SECRET_KEY"),
            },
        }):
            with Checkpoint() as cp:
                latest = cp.list(
                    task=run["start"].task
                )[0]
                print(latest)
                cp.load(
                    latest,
                    "test-checkpoints"
                )

        task = Task("TorchTuneFlow/8484/train/53673")
        with artifact_store_from(run=run, config={
            "client_params": {
                "aws_access_key_id": os.environ.get("MY_CUSTOM_ACCESS_KEY"),
                "aws_secret_access_key": os.environ.get("MY_CUSTOM_SECRET_KEY"),
            },
        }):
            load_model(
                task.data.model_ref,
                "test-models"
            )
        ```
    Parameters:
    ----------

    type: str
        The type of the datastore. Can be one of 's3', 'gcs', 'azure' or any other supported metaflow Datastore.

    config: dict or Callable
        Dictionary of configuration options for the datastore. The following keys are required:
        - root: The root path in the datastore where the data will be saved. (needs to be in the format expected by the datastore)
            - example: 's3://bucket-name/path/to/root'
            - example: 'gs://bucket-name/path/to/root'
            - example: 'https://myblockacc.blob.core.windows.net/metaflow/'
        - role_arn (optional): AWS IAM role to access s3 bucket (only when `type` is 's3')
        - session_vars (optional): AWS session variables to access s3 bucket (only when `type` is 's3')
        - client_params (optional): AWS client parameters to access s3 bucket (only when `type` is 's3')

    """

    name = "with_artifact_store"
    defaults = {
        "type": None,  # can be one of 's3', 'gcs', 'azure'
        "config": None,  # Dictionary of configuration options
    }

    def _export_context(self):
        _flw_deco = self
        _config = _flw_deco.attributes["config"]
        # Check if the config is a callable lambda
        if callable(_flw_deco.attributes["config"]):
            _config = _flw_deco.attributes["config"]()
        return {"type": _flw_deco.attributes["type"], "config": _config}

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        from .context import datastore_context

        cannot_be_none = ["type", "config"]
        for k in cannot_be_none:
            if self.attributes[k] is None:
                raise MetaflowException(
                    f"@{self.name} cannot have {k} be set to `None`."
                )
        _current_context = self._export_context()
        datastore_context.flow_init_context(_current_context)
        task_md = datastore_context.to_task_metadata()
        metadata.add_sticky_tags(
            sys_tags=["%s:%s" % (k, v) for k, v in task_md.items()]
        )


def set_datastore_context(flow, metadata, run_id, step_name, task_id, retry_count):
    from .context import datastore_context

    _flw_deco = None
    for decorators in flow._flow_decorators.values():
        for dec in decorators:
            if isinstance(dec, ArtifactStoreFlowDecorator):
                _flw_deco = dec
        else:
            continue

    if _flw_deco is None:
        # print("No datastore context found to set")
        return None

    _current_context = _flw_deco._export_context()
    datastore_context.flow_init_context(_current_context)

    task_md = datastore_context.to_task_metadata()
    entries = [
        MetaDatum(
            field=k,
            value=v,
            type="mf-checkpoint-artifact-store",
            tags=[
                "attempt_id:%s" % str(retry_count),
            ],
        )
        for k, v in task_md.items()
    ]
    metadata.register_metadata(run_id, step_name, task_id, entries)



from typing import Union, Optional, TYPE_CHECKING
from metaflow.exception import MetaflowException
import sys
from collections import namedtuple

# import context manager
from contextlib import contextmanager
import os
import json

ARTIFACT_STORE_CONFIG_ENV_VAR = "METAFLOW_CHECKPOINT_ARTIFACT_STORE_CONFIG"

if TYPE_CHECKING:
    import metaflow


def will_switch_context(func):
    """
    A decorator that wraps methods in ArtifactStoreContext.
    It takes the return value of the decorated method and passes it to
    self.switch_context(), then returns None.
    """

    def wrapper(self, *args, **kwargs):
        context = func(self, *args, **kwargs)
        self.switch_context(context)
        return None

    return wrapper


DatastoreContextValidation = namedtuple(
    "DatastoreContextInfo", ["is_valid", "needs_external_context", "context_mismatch"]
)


class UnresolvableDatastoreException(MetaflowException):
    pass


class ArtifactStoreContext:

    """
    This class will act as a singleton to switch the datastore so that
    Any checkpoint operations will be done in the correct datastore. This
    can be a useful way of short-circuting a datastore switch between runtime
    And post-runtime retrieval operations.
    """

    current_context = None  # dict

    _MF_DATASTORES = None

    @property
    def MF_DATASTORES(self):
        if self._MF_DATASTORES is None:
            from metaflow.plugins import DATASTORES

            self._MF_DATASTORES = DATASTORES
        return self._MF_DATASTORES

    def __init__(self):
        if ARTIFACT_STORE_CONFIG_ENV_VAR in os.environ:
            self.switch_context(json.loads(os.environ[ARTIFACT_STORE_CONFIG_ENV_VAR]))

    @will_switch_context
    def flow_init_context(self, context):
        if context is None:
            return None
        required_context_keys = ["type", "config"]
        required_config_keys = [
            "root",
        ]
        if not all([k in context for k in required_context_keys]):
            raise ValueError("Context does not have all required keys.")
        if not all([k in context["config"] for k in required_config_keys]):
            raise ValueError("Config does not have all required keys.")
        os.environ[ARTIFACT_STORE_CONFIG_ENV_VAR] = json.dumps(context)
        return context

    def switch_context(self, context):
        self.current_context = context

    def get(self):
        if self.current_context is None:
            return self.default()
        return self._from_current_context()

    def _from_current_context(self):
        # from metaflow.plugins import DATASTORES
        _type = self._runtime_ds_type(self.current_context["type"])
        _config = self.current_context["config"].copy()
        root = _config.pop("root")
        storage_impl = [d for d in self.MF_DATASTORES if d.TYPE == _type][0]
        return storage_impl(root, **_config)

    def _runtime_ds_type(self, ds_type):
        # Hack ; Ideally all datastores should have better
        # abstractions upstream. This will help handle the
        # case where we don't need to vendor the `s3-compatible`
        # datastore in the plugin
        if ds_type == "s3":
            return "s3-compatible"
        return ds_type

    def _task_metadata_ds_type(self, ds_type):
        # Hack ; Ideally all datastores should have better
        # abstractions upstream. This will help handle the
        # case where we don't need to vendor the `s3-compatible`
        # datastore in the plugin
        if ds_type == "s3-compatible":
            return "s3"
        return ds_type

    @will_switch_context
    def context_from_run(self, run: "metaflow.Run", config=None):
        md_keys = ["artifact-store-ds-type", "artifact-store-ds-root"]
        if not all([any([k in t for t in run.system_tags]) for k in md_keys]):
            # When we return None, we switch context to the default datastore.
            return None

        # extract artifact store type and root from system tags
        for t in run.system_tags:
            if t.startswith("artifact-store-ds-type:"):
                ds_type = t.split(":")[1]
                continue
            if t.startswith("artifact-store-ds-root:"):
                ds_root = ":".join(t.split(":")[1:])
                continue

        usable_context = {"type": ds_type, "config": {"root": ds_root}}
        if config is not None:
            usable_context["config"].update(config)

        return usable_context

    @will_switch_context
    def context_from_task(self, task: "metaflow.Task", config=None):
        metadata = task.metadata_dict
        md_keys = ["artifact-store-ds-type", "artifact-store-ds-root"]
        if not all([k in metadata for k in md_keys]):
            # When we return None, we switch context to the default datastore.
            return None

        ds_type = metadata["artifact-store-ds-type"]
        ds_root = metadata["artifact-store-ds-root"]
        usable_context = {"type": ds_type, "config": {"root": ds_root}}
        if config is not None:
            usable_context["config"].update(config)
        return usable_context

    def current_context_matches_task_metadata(self, task: "metaflow.Task"):
        metadata = task.metadata_dict
        md_keys = ["artifact-store-ds-type", "artifact-store-ds-root"]
        default_keys = ["ds-type", "ds-root"]
        task_ds_info = {"type": None, "root": None, "used_external_ds": False}
        if all([k in metadata for k in md_keys]):
            # This means that the task was run using an external artifact store.
            # so we can just return true.
            task_ds_info["type"] = metadata["artifact-store-ds-type"]
            task_ds_info["root"] = metadata["artifact-store-ds-root"]
            task_ds_info["used_external_ds"] = True
        elif all([k in metadata for k in default_keys]):
            # This means that the task was run using the default datastore.
            # so we can just return true.
            task_ds_info["type"] = metadata["ds-type"]
            task_ds_info["root"] = metadata["ds-root"]
        else:
            # We only raise an exeception because we are unable to resolve the datastore. This happens in the following situation:
            # The remote task has been deployed and didn't start execution and the user ctrl+c'd the job.
            # The means that metaflow's `step` command has not been able to register the `ds-type` and `ds-root` task metadata.
            # And in this scenarios even trying to access the checkpoints will make no difference. Since we would have essentially
            # written nothing.
            raise UnresolvableDatastoreException(
                "The datastore for the task was not found in the task metadata. This happens when the task didn't begin execution and "
                "it was killed by the user. Essentially no objects would have been written for this task by the "
                "`@checkpoint`/`@model`/`@huggingface_hub` decorators."
            )

        if self.current_context is not None:
            # This means that we are NOT in the default datastore context.
            # this means that user has explicitly set the datastore context in the flow code (via `@with_artifact_store`)
            # or via the `artifact_store_from` context manager.
            if not all(
                [self.current_context[k] == task_ds_info[k] for k in ["type", "root"]]
            ):
                # "TODO: Explain to the user that the execution context is setup to use an external datastore that is different from the task metadata. "
                # "And that they need to use the `artifact_store_from` context manager to configure the external datastore."
                return DatastoreContextValidation(
                    is_valid=False, needs_external_context=False, context_mismatch=True
                )
        else:
            if task_ds_info["used_external_ds"]:
                # "TODO: Explain to the user that they are using they need to use the `artifact_store_from` context manager to configure the external datastore "
                # "And that currently the client is only setup to access the default datastore."
                return DatastoreContextValidation(
                    is_valid=False, needs_external_context=True, context_mismatch=False
                )
        return DatastoreContextValidation(
            is_valid=True, needs_external_context=False, context_mismatch=False
        )

    def to_task_metadata(self):
        _ds = self.get()
        return {
            "artifact-store-ds-type": self._task_metadata_ds_type(_ds.TYPE),
            "artifact-store-ds-root": _ds.datastore_root,
        }

    def default(self):
        from metaflow.metaflow_config import DEFAULT_DATASTORE

        search_datastore = self._runtime_ds_type(DEFAULT_DATASTORE)

        storage_impl = [d for d in self.MF_DATASTORES if d.TYPE == search_datastore][0]
        return storage_impl(storage_impl.get_datastore_root_from_config(print))


datastore_context = ArtifactStoreContext()


@contextmanager
def artifact_store_from(run=None, task=None, config=None):
    """
    This context manager can be used to switch the artifact store
    for a block of code. This is useful when users maybe accessing
    checkpoints/models from a different datastore using the
    `@with_artifact_store` decorator.
    """
    if run is None and task is None:
        raise ValueError("Either run or task must be provided")
    if run is not None and task is not None:
        raise ValueError("Only one of run or task can be provided")

    try:
        if run is not None:
            datastore_context.context_from_run(run, config)
        else:
            datastore_context.context_from_task(task, config)
        yield
    finally:
        datastore_context.switch_context(None)



# THIS FILE IS ALSO USED (Figure where is the best place to put it.)
from typing import Union, Optional, TYPE_CHECKING
from metaflow import current
from metaflow.exception import MetaflowException
from .context import datastore_context

if TYPE_CHECKING:
    import metaflow


class UnresolvableDatastoreException(MetaflowException):
    pass


def init_datastorage_object():
    return datastore_context.get()


def resolve_storage_backend(pathspec: Union[str, "metaflow.Task"] = None):
    """
    This function is ONLY called when the user is calling `Checkpoint.list`.
    What happens when users call list:
        1. The task has completed so the list call is using the `_task_checkpoints` data-artifact to list the checkpoints.
        2. The task is still running or has crashed. This means that in order to list the checkpoints, we need access to the datastore.

    For case `1`, this code path wont even be called because the data-artifact is already a separate object.
    For case `2` is where this code path is important. But since we now expose a `artifact_store_from` context manager, we
    know pre-hand what the datastore needs to be and the datastore context has already been switched.

    SO in turn, one can make the argument that this function is not needed and we can just have a
    users do a `Checkpoint.list` under the context manager if they need to access objects in a different datastore.

    This is a respectible pattern since directly reading the task metadata and trying to do a list call is not a good pattern
    since the creds of metaflow default datastore might not be the same as the datastore the user wants to access.

    There is a function that can help verify if the datastore set in the task metadata
    is the same as the default datastore. If it is not, then we should shout warning messages
    to the user. In case the permissions are the same, nothing wrong happens, if they are not then
    user will have some hint in the logs to help them figure out the issue.
    """
    if isinstance(pathspec, str):
        from metaflow import Task

        task = Task(pathspec)
    else:
        task = pathspec
    validation = datastore_context.current_context_matches_task_metadata(task)
    return validation, datastore_context.get()


class FlowNotRunningException(MetaflowException):
    pass


def storage_backend_from_flow(flow: "metaflow.FlowSpec"):
    if not current.is_running_flow:
        raise FlowNotRunningException
    return datastore_context.get()
